//////////////////////////////////////////////////////////////////////////
///
/// @file ApiNet.h
///
/// @brief @brief Ring buffer description
///
/// @author ?
///
/// @brief Ring buffer description
///
/// Copyright (c) 2006 Parallels Inc.
/// All Rights Reserved.
/// http://www.parallels.com
///
//////////////////////////////////////////////////////////////////////////

#ifndef API_NET_H
#define API_NET_H

// We need some standard types definitions here
#include "ParallelsTypes.h"

// Packing structures with 1 byte
#ifdef _WIN_
#pragma pack(push, save_pack, 1)
#endif

// Ring buffer for packets maximum size
#define IO_NET_SIZE (256*1024)

/**
 * @brief
 *		Ring buffer for packet reception description
 *
 *		Struct of ring buffer:
 *		|<---------------------------------------- NET_BUFFER ------------------------------------------->|
 *		|                                                                                                 |
 *		|             |<--------------------------------- NET_BUFFER.aBuffer ---------------------------->|
 *		|             |                                                                                   |
 *		+-------------+--------------------------------+-----+--------------------------------+-----------+
 *		| Info about  |        Received packet         | ... |        Received packet         | Free      |
 *		| ring buffer.+---------------+----------------+ ... +---------------+----------------+           +
 *		| (aBuffer)   | PACKET_HEADER | Packet(data)   | ... | PACKET_HEADER | Packet(data)   |  area     |
 *		+-------------+---------------+----------------+-----+---------------+----------------+-----------+
 *		              ^
 *		              0
 */
typedef struct _NET_BUFFER
{
	// Packet(s) present into buffer
	UINT	bRecPacketPresent;
	// Packet(s) present in send buffer
	UINT	bSendPacketPresent;
	// Virtual cable is plugged
	UINT	bCableUnplugged;
	// Set to true when no space is available for sending packets
	UINT    bBufferFull;

	// Types of packets to accept
	// broad/multicasts always pass currently
	UINT	uRecFilter;
#define NET_FILTER_DIRECTED 1
#define NET_FILTER_BROADCAST 2
#define NET_FILTER_MULTICAST 4
#define NET_FILTER_PROMISCUOUS 8
#define NET_FILTER_ACCEPT_ALL (NET_FILTER_DIRECTED | NET_FILTER_BROADCAST | NET_FILTER_MULTICAST)
#define NET_FILTER_DISABLE_ADAPTER 0

	// Head of buffer (first allocated packet)
	UINT	uHead;
	// Tail of buffer
	UINT	uTail;
	// Head of buffer (first allocated packet)
	UINT	uSendHead;
	// Tail of buffer
	UINT	uSendTail;

	//
	// statistics
	//
	LONG64	bytes_in;  // bytes in
	LONG64  bytes_out; // bytes out
	LONG64 	pkts_in;   // packets in
	LONG64 	pkts_out;  // packets out
	LONG64	bcast_in;  // multicast packets in
	LONG64	mcast_in;  // multicast packets out
	LONG64	pkt_dropped_in; // number of in packet dropped due to buf overflow
	LONG64	pkt_err_in; // number of in packets dropped due to errors in packet
	LONG64	pkt_err_out; // number of out packets dropped due to errors in packet
	LONG64	send_buffer_notifications; // notifications to do immediate sends

	// align to 1Kb boundary here
	UCHAR	pad[1024 - 116];

	// data buffers allow 1.5Kb tailroom to ease wraparound
	UCHAR	aBuffer[IO_NET_SIZE + 1536];
	UCHAR	aSendBuffer[IO_NET_SIZE + 1536];
} STRUCT_NAME(NET_BUFFER);
typedef NET_BUFFER *PNET_BUFFER;

struct AssertIoNetSize {
	char assert[IO_NET_SIZE & (IO_NET_SIZE-1) ? -1 : 1];
};
struct AssertNetBufferStructure {
	char assert[sizeof(NET_BUFFER) != 4096 + IO_NET_SIZE * 2 ? -1 : 1];
};

/**
 * @brief
 *		Received Packet information
 */
typedef struct _NET_PACKET
{
	// eth header happens to be 14 bytes long
	// align IP payload on 16 byte boundary
	USHORT	uDataLen;
	// Packed data start
	UCHAR	aDataBuf[1];
} STRUCT_NAME(NET_PACKET);

struct AssertIPHeaderAlignment {
	char assert[(sizeof(NET_PACKET) - 1 + 14) & 15 ? -1 : 1];
};

// align packets to 16 byte boundary
#define FULL_PACKET_SIZE(uPacketDataLen) \
	((sizeof(NET_PACKET) - 1 + (uPacketDataLen) + 15) & ~15)

#define NET_GET_SND_PACKET(ptr, offset) ((NET_PACKET *) \
		(((NET_BUFFER *)ptr)->aSendBuffer + \
		(offset & (IO_NET_SIZE - 1))))

#define NET_GET_RCV_PACKET(ptr, offset) ((NET_PACKET *) \
		(((NET_BUFFER *)ptr)->aBuffer + \
		(offset & (IO_NET_SIZE - 1))))

// Packet packing restoring
#ifdef _WIN_
	#pragma pack(pop, save_pack)
#endif

#endif // API_NET_H

