/*-
 * Copyright (c) 2009, Alexander Nedotsukov
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice unmodified, this list of conditions, and the following
 *    disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */
/*-
 * This driver was ported from Parallels Linux Guest Tools
 * Copyright (C) 2008 Parallels Software Inc.
 * All Rights Reserved.
 * http://www.parallels.com
 */

#include <sys/cdefs.h>

#include <sys/param.h>
#include <sys/endian.h>
#include <sys/systm.h>
#include <sys/mbuf.h>
#include <sys/module.h>
#include <sys/lock.h>
#include <sys/mutex.h>
#include <sys/kernel.h>
#include <sys/socket.h>
#include <sys/sockio.h>

#include <net/if.h>
#include <net/if_arp.h>
#include <net/ethernet.h>
#include <net/if_media.h>
#include <net/if_types.h>

#include <net/bpf.h>

#include <machine/bus.h>
#include <machine/resource.h>
#include <sys/bus.h>
#include <sys/rman.h>

#include <dev/pci/pcireg.h>
#include <dev/pci/pcivar.h>

#include "if_pvmnetvar.h"

static int pvmnet_probe(device_t);
static int pvmnet_attach(device_t);
static int pvmnet_detach(device_t);

static void pvmnetintr(void *);
static int pvmnet_alloc_resources(device_t);
static void pvmnet_release_resources(device_t);

static void pvmnet_init(void *);
static void pvmnet_init_locked(struct pvmnet_softc *);
static int pvmnet_ioctl(struct ifnet *, u_long, caddr_t);
static void pvmnet_start(struct ifnet *);
static void pvmnet_start_locked(struct ifnet *);
static void pvmnet_stop(struct pvmnet_softc *);
static void pvmnet_setmulti(struct pvmnet_softc *);
static int pvmnet_encap(struct pvmnet_softc *, struct mbuf *);

static inline int
io_enable_interface(struct pvmnet_softc* sc)
{
	pvmnet_outl(sc, PRLETH_PORT_COMMAND, PRLETH_COMMAND_PRLETH);
	pvmnet_outw(sc, PRLETH_PORT_COMMAND, PRLETH_COMMAND_ENABLE);
	return 0;
}

static inline void
io_disable_interface(struct pvmnet_softc *sc)
{
	pvmnet_outw(sc, PRLETH_PORT_COMMAND, PRLETH_COMMAND_DISABLE);
	pvmnet_outl(sc, PRLETH_PORT_COMMAND, PRLETH_COMMAND_RTL8029);
}

static inline void
io_get_mac_address(struct pvmnet_softc *sc, u_char *mac, unsigned int len)
{
	pvmnet_outw(sc, PRLETH_PORT_COMMAND, PRLETH_COMMAND_GET_ADDRESS);
	pvmnet_insb(sc, PRLETH_PORT_DATA, mac, len);
}

static inline int
io_write_data(struct pvmnet_softc *sc, struct mbuf *m)
{
	unsigned int size, fullsize;
	unsigned int end, start;
	NET_PACKET *packet;
	bus_size_t dst;

	start = pvmnet_readl(sc, offsetof(NET_BUFFER, uSendTail));
	end =  pvmnet_readl(sc, offsetof(NET_BUFFER, uSendHead));

	size = m_length(m, NULL);
	fullsize = FULL_PACKET_SIZE(size);
	if (fullsize > (IO_NET_SIZE - (start - end))) {
		/* Buffer is full */
		return -1;
	}

	packet = NET_GET_SND_PACKET(0, start);
	pvmnet_writew(sc, (bus_size_t)&packet->uDataLen, size);
	dst = (bus_size_t)packet->aDataBuf;
	for (; m != NULL; m = m->m_next) {
		pvmnet_write(sc, dst, mtod(m, uint8_t*), m->m_len);
		dst += m->m_len;
	}

	pvmnet_writel(sc, offsetof(NET_BUFFER, uSendTail), start + fullsize);
	pvmnet_writel(sc, offsetof(NET_BUFFER, bSendPacketPresent), 1);

	return (0);
}

static inline void
io_notify_sndbuf_full(struct pvmnet_softc *sc)
{
	pvmnet_outl(sc, PRLETH_PORT_COMMAND, PRLETH_COMMAND_BUFOVERFLOW);
}

static inline void
io_interrupts(struct pvmnet_softc *sc, int enable)
{
	unsigned int mask;

	if (enable)
		mask =  PRLETH_STATUS_RECVD
			| PRLETH_STATUS_CABLE
			| PRLETH_STATUS_SPACE;
	else
		mask = 0;

	pvmnet_outw(sc, PRLETH_PORT_MASK, mask);
}

static inline unsigned int
io_get_rcv_offset(struct pvmnet_softc *sc)
{
	return pvmnet_readl(sc, offsetof(NET_BUFFER, uHead));
}

static inline void
io_move_rcv_offset(struct pvmnet_softc *sc, unsigned int offset)
{
	pvmnet_outl(sc, PRLETH_PORT_MOVE_RECV_HEAD, offset);
}

static inline unsigned int
io_get_packet_size(struct pvmnet_softc *sc, unsigned int offset)
{
	unsigned int end;
	NET_PACKET *packet;

	end = pvmnet_readl(sc, offsetof(NET_BUFFER, uTail));

	if (offset == end)
		return 0;       /* No more messages */

	packet = NET_GET_RCV_PACKET(0, offset);
	return pvmnet_readw(sc, (bus_size_t)&packet->uDataLen);
}

static inline void
io_read_data(struct pvmnet_softc *sc, unsigned int offset, struct mbuf *m)
{
	NET_PACKET *packet;

	packet = NET_GET_RCV_PACKET(0, offset);
	pvmnet_read(sc, (bus_size_t)packet->aDataBuf, mtod(m, uint8_t*), m->m_len);
}

static inline unsigned int
io_get_status(struct pvmnet_softc *sc)
{
	return pvmnet_inw(sc, PRLETH_PORT_STATUS);
}

static inline void
io_multicast(struct pvmnet_softc *sc, int enable)
{
	unsigned int cmd;

	if (enable)
		cmd = PRLETH_COMMAND_PASS_MULTICAST;
	else
		cmd = PRLETH_COMMAND_FILTER_MULTICAST;

	pvmnet_outw(sc, PRLETH_PORT_COMMAND, cmd);
}

static inline void
io_promisc(struct pvmnet_softc *sc, int enable)
{
	unsigned int cmd;

      	if (enable)
		cmd = PRLETH_COMMAND_SET_PROMISCUOUS;
	else
		cmd = PRLETH_COMMAND_CLR_PROMISCUOUS;

	pvmnet_outw(sc, PRLETH_PORT_COMMAND, cmd);
}

MODULE_DEPEND(pvmnet, pci, 1, 1, 1);
MODULE_DEPEND(pvmnet, ether, 1, 1, 1);

static device_method_t pvmnet_methods[] = {
	/* Device interface */
	DEVMETHOD(device_probe,		pvmnet_probe),
	DEVMETHOD(device_attach,	pvmnet_attach),
	DEVMETHOD(device_detach,	pvmnet_detach),

	/* bus interface */
	DEVMETHOD(bus_print_child,	bus_generic_print_child),
	DEVMETHOD(bus_driver_added,	bus_generic_driver_added),

	{ 0, 0 }
};

static driver_t pvmnet_driver = {
	"pvmnet",
	pvmnet_methods,
	sizeof(struct pvmnet_softc)
};

static devclass_t pvmnet_devclass;

DRIVER_MODULE(pvmnet, pci, pvmnet_driver, pvmnet_devclass, 0, 0);

static int
pvmnet_probe(device_t dev)
{
	uint16_t devid, vendor;
	uint16_t revid, sdevid;

	vendor = pci_get_vendor(dev);
	devid = pci_get_device(dev);
	revid = pci_get_revid(dev);
	sdevid = pci_get_subdevice(dev);

	if (vendor != 0x10ec || devid != 0x8029)
		return (ENXIO);
	device_set_desc(dev, "Parallels VM network driver");
	return (BUS_PROBE_VENDOR);
}

static int
pvmnet_attach(device_t dev)
{
	struct pvmnet_softc *sc = device_get_softc(dev);
	int error;
	struct ifnet *ifp;

	sc->dev = dev;
	PVMNET_LOCK_INIT(sc);

	pci_enable_busmaster(dev);
	pci_enable_io(dev, SYS_RES_IOPORT);
	pci_enable_io(dev, SYS_RES_MEMORY);

	error = pvmnet_alloc_resources(dev);
	if (error) {
		pvmnet_release_resources(dev);
		return (error);
	}
	io_enable_interface(sc);
	io_get_mac_address(sc, sc->enaddr, sizeof(sc->enaddr));

	ifp = sc->ifp = if_alloc(IFT_ETHER);
	if (ifp == NULL) {
		pvmnet_release_resources(dev);
		return (ENOSPC);
	}
	ifp->if_softc = sc;
	if_initname(ifp, device_get_name(dev), device_get_unit(dev));
	ifp->if_mtu = ETHERMTU;
	ifp->if_flags = IFF_BROADCAST | IFF_SIMPLEX | IFF_MULTICAST;
	ifp->if_start = pvmnet_start;
	ifp->if_ioctl = pvmnet_ioctl;
	ifp->if_init = pvmnet_init;
	ifp->if_baudrate = 1000000000;
	IFQ_SET_MAXLEN(&ifp->if_snd, IFQ_MAXLEN);
	ifp->if_snd.ifq_drv_maxlen = IFQ_MAXLEN;
	IFQ_SET_READY(&ifp->if_snd);
	ether_ifattach(ifp, sc->enaddr);

	return (0);
}

static int
pvmnet_detach(device_t dev)
{
	struct pvmnet_softc *sc = device_get_softc(dev);
	struct ifnet *ifp = sc->ifp;

	PVMNET_LOCK(sc);
	pvmnet_stop(sc);
	PVMNET_UNLOCK(sc);
	ether_ifdetach(ifp);
	if_free(ifp);
	io_disable_interface(sc);
	pvmnet_release_resources(dev);
	PVMNET_LOCK_DESTROY(sc);

	return (0);
}

static void
pvmnetintr(void *arg)
{
	struct pvmnet_softc *sc = (struct pvmnet_softc *)arg;
	struct ifnet *ifp;
	int status;
	unsigned int offset;
	unsigned int size;
	struct mbuf *m;

	PVMNET_LOCK(sc);
	ifp = sc->ifp;
	if (!(ifp->if_flags & IFF_UP)) {
		PVMNET_UNLOCK(sc);
		return;
	}
	status = io_get_status(sc);
	if (!(status & (PRLETH_STATUS_RECVD | PRLETH_STATUS_SPACE | PRLETH_STATUS_CABLE))) {
		/* No interrupt: shared IRQs cause this */
		PVMNET_UNLOCK(sc);
		return;
	}
        if (status & PRLETH_STATUS_CABLE) {
                if (status & PRLETH_STATUS_CONNECTED)
			if_link_state_change(ifp, LINK_STATE_UP);	
		else
			if_link_state_change(ifp, LINK_STATE_DOWN);	
                
        }
	if (status & PRLETH_STATUS_SPACE) {
		ifp->if_drv_flags &= ~IFF_DRV_OACTIVE;
		pvmnet_start_locked(ifp);
	}
	if (status & PRLETH_STATUS_RECVD) {
		for (offset = io_get_rcv_offset(sc);
		    (size = io_get_packet_size(sc, offset)); offset += FULL_PACKET_SIZE(size)) {
			if (size > ETHER_MAX_LEN) {
				ifp->if_ierrors++;
				continue;
			}
			m = m_getcl(M_DONTWAIT, MT_DATA, M_PKTHDR);
			if (m == NULL) {
				ifp->if_ierrors++;
				continue;
			}
			m->m_pkthdr.rcvif = ifp;
			m->m_pkthdr.len = m->m_len = MCLBYTES;
			m_adj(m, ETHER_ALIGN);

			m->m_pkthdr.len = m->m_len = size;
			io_read_data(sc, offset, m);

			ifp->if_ipackets++;

			PVMNET_UNLOCK(sc);
			(*ifp->if_input)(ifp, m);
			PVMNET_LOCK(sc);
		}
		io_move_rcv_offset(sc, offset); /* Enables interrupts */
		goto exit;
	}
	io_interrupts(sc, TRUE);
exit:
	PVMNET_UNLOCK(sc);
}

static int pvmnet_alloc_resources(device_t dev)
{
	struct pvmnet_softc *sc = device_get_softc(dev);
	struct resource *res;
	int rid;
	int error;

	rid = PCIR_BAR(0);
	res = bus_alloc_resource_any(dev, SYS_RES_IOPORT, &rid, RF_ACTIVE);
	if (!res)
		return ENXIO;
	sc->port_rid = rid;
	sc->port_res = res;
	sc->port_bst = rman_get_bustag(res);
	sc->port_bsh = rman_get_bushandle(res);

	rid = PCIR_BAR(1);
	res = bus_alloc_resource_any(dev, SYS_RES_MEMORY, &rid, RF_ACTIVE);
	if (!res)
		return ENXIO;
	sc->mem_rid = rid;
	sc->mem_res = res;
	sc->mem_bst = rman_get_bustag(res);
	sc->mem_bsh = rman_get_bushandle(res);

	rid = 0;
	res = bus_alloc_resource_any(dev, SYS_RES_IRQ, &rid, RF_ACTIVE | RF_SHAREABLE);
	if (!res)
		return ENXIO;
	sc->irq_res = res;
	sc->irq_rid = rid;

	error = bus_setup_intr(dev, sc->irq_res, INTR_TYPE_NET | INTR_MPSAFE,
		NULL, pvmnetintr, sc, &sc->irq_handle);
	if (!res)
		return ENXIO;
	return (0);
}

static void pvmnet_release_resources(device_t dev)
{
	struct pvmnet_softc *sc = device_get_softc(dev);

	if (sc->irq_handle)
		bus_teardown_intr(dev, sc->irq_res, sc->irq_handle);
	if (sc->irq_res)
		bus_release_resource(dev, SYS_RES_IRQ, sc->irq_rid, sc->irq_res);
	if (sc->mem_res)
		bus_release_resource(dev, SYS_RES_MEMORY, sc->mem_rid, sc->mem_res);
	if (sc->port_res)
		bus_release_resource(dev, SYS_RES_IOPORT, sc->port_rid, sc->port_res);
}

static void
pvmnet_init(void *xsc)
{
	struct pvmnet_softc *sc = xsc;

	PVMNET_ASSERT_UNLOCKED(sc);
	PVMNET_LOCK(sc);
	pvmnet_init_locked(sc);
	PVMNET_UNLOCK(sc);
}

static void
pvmnet_init_locked(struct pvmnet_softc *sc)
{
	struct ifnet *ifp = sc->ifp;
	int status;

	PVMNET_ASSERT_LOCKED(sc);

	status = io_get_status(sc);

	pvmnet_setmulti(sc);

	io_interrupts(sc, TRUE);

	if (status & PRLETH_STATUS_CONNECTED)
		if_link_state_change(ifp, LINK_STATE_UP);	
	else
		if_link_state_change(ifp, LINK_STATE_DOWN);	

	ifp->if_drv_flags |= IFF_DRV_RUNNING;
	ifp->if_drv_flags &= ~IFF_DRV_OACTIVE;

	pvmnet_start_locked(ifp);
}

static int
pvmnet_ioctl(struct ifnet *ifp, u_long command, caddr_t data)
{
	struct pvmnet_softc *sc = ifp->if_softc;
	int error;

	switch (command) {
	case SIOCSIFFLAGS:
		PVMNET_LOCK(sc);
		if (ifp->if_flags & IFF_UP)
			pvmnet_init_locked(sc);
		else if (ifp->if_drv_flags & IFF_DRV_RUNNING)
			pvmnet_stop(sc);
		PVMNET_UNLOCK(sc);
		error = 0;
		break;
	case SIOCADDMULTI:
	case SIOCDELMULTI:
		PVMNET_LOCK(sc);
		pvmnet_setmulti(sc);
		PVMNET_UNLOCK(sc);
		error = 0;
		break;
	default:
		error = ether_ioctl(ifp, command, data);
	}

	return (error);
}

static void
pvmnet_start(struct ifnet *ifp)
{
	struct pvmnet_softc *sc = ifp->if_softc;

	PVMNET_ASSERT_UNLOCKED(sc);
	PVMNET_LOCK(sc);
	pvmnet_start_locked(ifp);
	PVMNET_UNLOCK(sc);
}

static void
pvmnet_start_locked(struct ifnet *ifp)
{
	struct pvmnet_softc *sc = ifp->if_softc;
	struct mbuf *m;

	PVMNET_ASSERT_LOCKED(sc);
	while (!IFQ_DRV_IS_EMPTY(&ifp->if_snd)) {
		IFQ_DRV_DEQUEUE(&ifp->if_snd, m);
		if (pvmnet_encap(sc, m)) {
			IFQ_DRV_PREPEND(&ifp->if_snd, m);
			ifp->if_drv_flags |= IFF_DRV_OACTIVE;
			io_notify_sndbuf_full(sc);
			break;
		}
		ifp->if_opackets++;
		BPF_MTAP(ifp, m);
		m_freem(m);
	}
}
static void
pvmnet_stop(struct pvmnet_softc *sc)
{
	PVMNET_ASSERT_LOCKED(sc);

	io_interrupts(sc, FALSE);
}

static void pvmnet_setmulti(struct pvmnet_softc * sc)
{
	struct ifnet *ifp = sc->ifp;

	PVMNET_ASSERT_LOCKED(sc);

	if (ifp->if_flags & IFF_PROMISC)
		io_promisc(sc, 1);
	else
		io_promisc(sc, 0);
	if (ifp->if_flags & IFF_ALLMULTI)
		io_multicast(sc, 1);
	else
		io_multicast(sc, 0);
}

static int
pvmnet_encap(struct pvmnet_softc *sc, struct mbuf *m)
{
	return io_write_data(sc, m);
}
